#include "BitmapParser.h"
#include "animation.h"

void ReorderSegment(byte *segment16x16)
{
    int x, y, dx, dy;
    int counter = 0;
    byte segmentCopy[128];

    for (dx = 0; dx < 8; dx += 4)
        for (dy = 0; dy < 16; dy += 8)
            for (y = 0; y < 8; y++)
                for (x = 0; x < 4; x++, counter++)
                {
                    segmentCopy[counter] = segment16x16[(y+dy)*8+(x+dx)];

                    #ifdef DEBUG
                    if (counter >= 128 || ((y+dy)*8+(x+dx))  >= 128)
                    {
                        printf("Grand Error #666\n");
                        WaitKey();
                    }
                    #endif
                }


    for (counter = 0; counter < 128; counter++)
        segment16x16[counter] = segmentCopy[counter];
}

void LoadSpriteGraphics(char *bitmapName, char *dataName)
{
    short id;
    byte *bitmapData;
    int rastersize;
    int x, y;
    int fdBitmap = OpenFile(bitmapName, 0);
    int fdData = OpenFile(dataName, 0);
    short animation, frame, animations, frames, animationPose;
    short gameObj, gameObjs;
    BitmapHeader header;
    GraphicLocation location;

    ReadFile(fdBitmap,(void *)&header,sizeof(BitmapHeader));
    header.width = EndianSwap32(header.width);
    header.height = EndianSwap32(header.height);

    rastersize = header.width*header.height*3;
    _dos_seek(fdBitmap, EndianSwap32(header.rasterOffset), 0);
    bitmapData = AllocateMemory(-rastersize);
    ReadFile(fdBitmap, bitmapData, rastersize);

    Wipe16(SPRITE_PALLET_ADR,COLOUR_TRANSPARANT,256);
    SPRITE_PALLET_ADR[0] = 0;

    //Here we load information about the number of graphics objects
    //and their location in the global graphic BMP.
	_dos_seek(fdData, 0, 0);
    animations = FetchINT16(fdData);

	for (animation = 0, id = 0; animation < animations; animation++)
    {
        frames = FetchINT16(fdData);

        //setup animation
        for (frame = 0; frame < frames; frame++, id++)
        {
            ReadFile(fdData, (void *)&location, sizeof(GraphicLocation));
            ParseSprite(bitmapData, header.width, header.height, location.x, location.y, location.dx, location.dy, &gameSpritePool[id], (unsigned short *)SPRITE_PALLET_ADR);
            //insert graphic frame pointer into animation
            animationPool[animation].frame[frame] = &gameSpritePool[id];
        }

        animationPool[animation].type = ANIMATION_TYPE_LOOP;
        animationPool[animation].numberOfFrames = frames;
    }

    //Now that animations have been loaded it is time to load the game objects that use the animations
    gameObjs = FetchINT16(fdData);

    for (gameObj = 0; gameObj < gameObjs; gameObj++)
    {
        Wipe32((void *)gameObjectPool[gameObj].animations, 0, MAX_OBJECT_ANIMATIONS);

        gameObjectPool[gameObj].type = FetchINT16(fdData);
        gameObjectPool[gameObj].width = FetchINT16(fdData);
        gameObjectPool[gameObj].height = FetchINT16(fdData);
        gameObjectPool[gameObj].horizontalSpriteOffset = FetchINT16(fdData);
        gameObjectPool[gameObj].verticalSpriteOffset = FetchINT16(fdData);

        animations = FetchINT16(fdData);

        for (animation = 0; animation < animations; animation++)
        {
            //Animation pose defines which situation the animation is for such as walking, being hit, attacking etc.
            animationPose = FetchINT16(fdData);
            //ID here is just an index in the animation pool, pointing out where the animation is stored.
            id = FetchINT16(fdData);
            gameObjectPool[gameObj].animations[animationPose] = &animationPool[id];
        }
    }

	//Get location of tiles in the bitmap
	ReadFile(fdData, (void *)&location, sizeof(GraphicLocation));
	ParseTiles((void*)bitmapData, header.width, header.height, location.x, location.y, location.dx, location.dy, (unsigned short *)GRAPHIC_PALLET_ADR);

    _dos_close(fdBitmap);
    _dos_close(fdData);

    //remember to free bitmapData allocation
    //_dos_mfree(bitmapData);
}

//Find index of a particular colour in a pallet, returns -1 if not found.
int LocateColour(unsigned short *pallet, unsigned short colour)
{
    int entry;

    for (entry = 1; entry < 16; entry++)
        if (pallet[entry] == colour)
            return entry;

    return -1;
}

//This function changes any pallet index that needs to be changed in sprite data
//because of earlier pallet mergers.
void RemapPalletEntries(SpriteSegment *spriteSegment, short *remap)
{
    bool lowOctRemapped;
    bool highOctRemapped;
    byte lowOct, highOct;
    int index, entry;

    for (entry = 0; entry < 128; entry++)
    {
        lowOctRemapped = false;
        highOctRemapped = false;
        for (index = 1; index < 16; index++)
        {
            //if remap index is not used or maps to the same value, skip
            if (remap[index] == -1 || remap[index] == index)
                continue;

            lowOct = spriteSegment->segment[entry]&0x0F;
            highOct = (spriteSegment->segment[entry]&0xF0)>>4;

            if (!lowOctRemapped && lowOct == index)
            {
                lowOct = remap[index];
                lowOctRemapped = true;
            }

            if (!highOctRemapped && highOct == index)
            {
                highOct = remap[index];
                highOctRemapped = true;
            }

            spriteSegment->segment[entry] = (highOct << 4)|lowOct;

            if (lowOctRemapped && highOctRemapped)
                break;
        }
    }
}

//This fucntion takes a pallet and attemps to insert it into the global pallet
//if possibel it will attempt to merge the provided pallet with an existing one
//but this action might rearrange some pallet index values which then need to be
//corrected in the sprite data.
int InsertPallet(unsigned short *globalPallet, SpriteSegment *spriteSegment, unsigned short *localPallet)
{
    int entry;
    int location;
    int palletID;
    int localPalletColoursUsed;
    int globalPalletColoursUsed;
    int coloursMissing;
    short remap[16];
    unsigned short missingColours[16];

    //Find out how many colours there are used in the local pallet
    localPalletColoursUsed = LocateColour(localPallet, COLOUR_TRANSPARANT);
    if (localPalletColoursUsed == -1)
        localPalletColoursUsed = 16;

    //There are no colours in the pallet (should be avoided when possible)
    if (localPalletColoursUsed == 1)
    {
        spriteSegment->palletID = 0;
        return 0;
    }

    for (palletID = 0; palletID < 16; palletID++)
    {
        //Find out how many colours there are used in the current global pallet
        globalPalletColoursUsed = LocateColour(&globalPallet[palletID * 16], COLOUR_TRANSPARANT);
        if (globalPalletColoursUsed == -1)
            globalPalletColoursUsed = 16;

        //We have reached the end of occupied pallets
        if (globalPalletColoursUsed == 1)
            break;

        //Current global pallet is used, time to see if its colours matches the desired local pallet
        coloursMissing = 0;
        Wipe16(remap, -1, 16);
        Wipe16(missingColours, COLOUR_TRANSPARANT, 16);

        for (entry = 1; entry < localPalletColoursUsed; entry++)
        {
            location = LocateColour(&globalPallet[palletID*16], localPallet[entry]);

            if (location == -1)
            {
                missingColours[coloursMissing] = localPallet[entry];
                coloursMissing++;
            }
            else
                remap[entry] = location;
        }

        //The current global pallet does not have all the colours we need and it is already full
        if (globalPalletColoursUsed == 16 && coloursMissing > 0)
            continue;

        //The current global pallet has all the colours we need.
        if (coloursMissing == 0)
        {
            RemapPalletEntries(spriteSegment, remap);

            //That is it, we are done!
            spriteSegment->palletID = palletID;
            return 0;
        }
        //We found a partially full pallet with room for the colours we need
        if (coloursMissing <= (16 - globalPalletColoursUsed))
        {
            //Fill out the vacant pallet entries with the desired colours
            //and save remapped index for colour in the new pallet.
            for (entry = globalPalletColoursUsed, location = 0; entry < globalPalletColoursUsed+coloursMissing; entry++, location++)
            {
                globalPallet[(palletID*16)+entry] = missingColours[location];
                remap[LocateColour(localPallet, missingColours[location])] = entry;
            }

            //Now all we need to do is remap the pallet entries in the sprite data
            //to the new indexes.
            RemapPalletEntries(spriteSegment, remap);

            //That is it, we are done!
            spriteSegment->palletID = palletID;

            return 0;
        }
    }

    //No pallet entry matched our colours, so we need to fill out an empty entry.
    if (palletID < 16)
        for (entry = 1; entry < 16; entry++)
            globalPallet[(palletID*16)+entry] = localPallet[entry];
    #ifdef DEBUG
    else
    {
        //No pallet entry matched and there are no free spaces left (this should not happen)
        printf("No room for pallet Error!\n");
        WaitKey();
        return-1;

    }
    #endif
    spriteSegment->palletID = palletID;

    return 0;
}

//This function parses a sprite of variable size out of a BMP image.
//The sprite is parsed in 16x16 chunks each with their own pallet.
void ParseSprite(byte *imageData, int imageWidth, int imageHeight, int x, int y, int spriteWidth, int spriteHeight, GameSprite *sprite, unsigned short *sharedSpritePallet)
{
    short vs;
    short hs;
    short vr;
    short hr;
    short hSegments = spriteWidth / 16;
    short vSegments = spriteHeight / 16;
    short hRemainder = spriteWidth % 16;
    short vRemainder = spriteHeight % 16;
    short dx, dy;
    int bitmapOffset;
    short segmentOffset;
    byte entry;
    unsigned short smallColour;
    unsigned short r,g,b;
    unsigned short pallet[16];
    //byte packedPixel;
    //byte pixelCounter;
    SpriteSegment *spriteSegment;

    // We need to account for the total sprite size not being divisable by 16.
    if (vRemainder != 0)
        vSegments++;
    else
        vRemainder = 16;

    if (hRemainder != 0)
        hSegments++;
    else
        hRemainder = 16;

    sprite->verticalSegments = vSegments;
    sprite->horizontalSegments = hSegments;

	#ifdef DEBUG
	if (vSegments*hSegments>= MAX_SPRITE_SEGMENTS)
	{
		printf("Error: Sprite has too many segments!\n");
		WaitKey();
	}
    #endif

    for (entry = 0; entry < (vSegments*hSegments); entry++)
        sprite->segments[entry] = AllocateMemory(sizeof(SpriteSegment));

    for (vs = 0; vs < vSegments; vs++)
    {
        //Each sprite segment is 16 pixels high
        //but we need to account for partial segments
        if (vs == vSegments-1)
            vr = vRemainder;
        else
            vr = 16;

        for (hs = 0; hs < hSegments; hs++)
        {
            //Each sprite segment is 16 pixels wide
            //but we need to account for a partial last segment
            if (hs == hSegments-1)
                hr = hRemainder;
            else
                hr = 16;

            spriteSegment = sprite->segments[(vs * hSegments) + hs];
            spriteSegment->pcgIndex = VACANT_SLOT;

            Wipe32((void *)spriteSegment->segment,0x00000000,32);
            Wipe16(pallet,COLOUR_TRANSPARANT,16);
            #ifdef DEBUG
            if ((vs * hSegments) + hs >= hSegments*vSegments)
            {
                printf("Grand Error #1\n");
                WaitKey();
            }
            #endif
            for (dy = 0; dy < vr; dy++)
            {
                // The address of a 16 pixel sprite segment in the Bitmap data
                bitmapOffset = (((imageHeight-1)-(y+dy+vs*16)) * imageWidth) + x+(hs*16);

                // Start with a empty pallet for each segment
                for (dx = 0; dx < hr; dx++)
                {
                    #ifdef DEBUG
                    if ((bitmapOffset + dx)*3+2 >= (imageWidth*imageHeight*3) || bitmapOffset < 0)
                    {
                        printf("Grand Error #2\n");
                        WaitKey();
                    }
                    #endif
                    // miltiplied by threee because each bitmap pixel is 3 bytes.
                    b = imageData[(bitmapOffset + dx)*3];
                    g = imageData[(bitmapOffset + dx)*3+1];
                    r = imageData[(bitmapOffset + dx)*3+2];

                    // We need to convert each colour component from 8 to 5 bit
                    r = (r * 31)/255;
                    g = (g * 31)/255;
                    b = (b * 31)/255;
                    smallColour = (g<<11) | (r<<6) | (b<<1);

                    if (smallColour == COLOUR_TRANSPARANT)
                        entry = 0;
                    else
                    {
                    	//Insert current colour into the temporary segment pallet
                    	//but only if the colour is not already in the pallet
                        for (entry = 1; entry < 16; entry++)
                        {
                            if (pallet[entry] == smallColour)
                                break;

                            if (pallet[entry] == COLOUR_TRANSPARANT)
                            {
                                pallet[entry] = smallColour;
                                break;
                            }

                            #ifdef DEBUG
                            // More than 16 unique colours in segment!
                            if (entry == 15)
                            {
                                printf("Too many colours!\n");
                                WaitKey();
                            }
                            #endif
                        }
                    }

					segmentOffset = ((dy << 4) + dx)>>1;
                    //Each byte contains two pixels (pallet entries)
                    if (dx%2==0)
						spriteSegment->segment[segmentOffset] = (entry&0x0F) << 4;
                    else
                    	spriteSegment->segment[segmentOffset] |= entry&0x0F;

                    #ifdef DEBUG
                    if (segmentOffset >= 128)
                    {
                        printf("Grand Error #3\n");
                        WaitKey();
                    }
                    #endif
                }
            }

			//Make the pallet of this sprite segment part of the global pallet
            InsertPallet(sharedSpritePallet, spriteSegment, pallet);
			//Change the 16x16 segments pixel arrangement into the particular 4 x (8x8)
			//arrangement that X68000 prefers
            ReorderSegment(spriteSegment->segment);
        }
    }
}

//This function parses a sprite of variable size out of a BMP image.
//The sprite is parsed in 16x16 chunks each with their own pallet.
void ParseTiles(byte *imageData, short imageWidth, short imageHeight, short x, short y, short horizontalTiles, short verticalTiles, unsigned short *sharedPallet)
{
    short vt;
    short ht;
    short dx, dy;
    int bitmapOffset;
    short tileOffset;
    byte entry;
    unsigned short smallColour;
    unsigned short r,g,b;
    GraphicTile *graphicTile;

	#ifdef DEBUG
	if (verticalTiles*horizontalTiles>= MAX_GRAPHIC_TILE_POOL_SIZE)
	{
		printf("Error: Too many tiles!\n");
		WaitKey();
	}
    #endif

//    for (entry = 0; entry < (vSegments*hSegments); entry++)
//        sprite->segments[entry] = AllocateMemory(sizeof(SpriteSegment));

	Wipe16(sharedPallet,COLOUR_TRANSPARANT,256);
	sharedPallet[0] = 0;

    for (vt = 0; vt < verticalTiles; vt++)
    {
        for (ht = 0; ht < horizontalTiles; ht++)
        {
            graphicTile = &graphicTilePool[(vt * horizontalTiles) + ht];

            Wipe32((void *)graphicTile->tile,0x00000000,64);
            #ifdef DEBUG
            if ((vt * horizontalTiles) + ht >= horizontalTiles*verticalTiles)
            {
                printf("Grand Error #2-1\n");
                WaitKey();
            }
            #endif
            for (dy = 0; dy < 16; dy++)
            {
                // The address of a 16 pixel sprite segment in the Bitmap data
                bitmapOffset = (((imageHeight-1)-(y+dy+vt*16)) * imageWidth)+x+(ht*16);

                // Start with a empty pallet for each segment
                for (dx = 0; dx < 16; dx++)
                {
                    #ifdef DEBUG
                    if ((bitmapOffset + dx)*3+2 >= (imageWidth*imageHeight*3) || bitmapOffset < 0)
                    {
						printf("horizontalTiles: %i\n", horizontalTiles);
						printf("verticalTiles: %i\n", verticalTiles);
						printf("x: %i\n", x);
						printf("y: %i\n", y);
						printf("dx: %i\n", dx);
						printf("dy: %i\n", dy);
						printf("vt: %i\n", vt);
						printf("ht: %i\n", ht);
						printf("imageWidth: %i\n", imageWidth);
						printf("imageHeight: %i\n", imageHeight);
						printf("bitmapOffset: %i\n", bitmapOffset);
                        printf("Grand Error #2-2\n");
                        WaitKey();
                    }
                    #endif
                    // miltiplied by threee because each bitmap pixel is 3 bytes.
                    b = imageData[(bitmapOffset + dx)*3];
                    g = imageData[(bitmapOffset + dx)*3+1];
                    r = imageData[(bitmapOffset + dx)*3+2];

                    // We need to convert each colour component from 8 to 5 bit
                    r = (r * 31)/255;
                    g = (g * 31)/255;
                    b = (b * 31)/255;
                    smallColour = (g<<11) | (r<<6) | (b<<1);

                    if (smallColour == COLOUR_TRANSPARANT)
                        entry = 0;
                    else
                    {
                    	//Insert current colour into the temporary segment pallet
                    	//but only if the colour is not already in the pallet
                        for (entry = 1; entry < 256; entry++)
                        {
                            if (sharedPallet[entry] == smallColour)
                                break;

                            if (sharedPallet[entry] == COLOUR_TRANSPARANT)
                            {
                                sharedPallet[entry] = smallColour;
                                break;
                            }

                            #ifdef DEBUG
                            // More than 256 unique colours in tiles!
                            if (entry == 256)
                            {
                                printf("Too many colours in tiles!\n");
                                WaitKey();
                            }
                            #endif
                        }
                    }

					tileOffset = ((dy << 4) + dx);
                    //Each byte contains one pixel (pallet entry)
					graphicTile->tile[tileOffset] = entry;

                    #ifdef DEBUG
                    if (tileOffset >= 256)
                    {
                        printf("Grand Error #2-3\n");
                        WaitKey();
                    }
                    #endif
                }
            }
        }
    }
}
