#include "GameLoop.h"
#include "Camera.h"
#include "GameObject.h"
#include "AI.h"
#include "Animation.h"
#include "Renderer.h"

void GameLoop()
{
	bool descending;
    short spawn, x, y, lx, ly;
    SpawnedGameObject *obj;

	wait_for_vblank();
	SetLayers(LAYER_NONE);

    if (camera.moved)
    {
        camera.moved = false;

        if (levelSurface.enabled)
			RenderSurface(&levelSurface, false);

        //if (backgroundSurface.enabled)
			RenderSurface(&backgroundSurface, false);
    }

    //Update all spawned objects
    for (spawn = 0; spawn < MAX_SPAWNED_GAMEOBJECT_POOL_SIZE; spawn++)
    {
    	obj = &spawnedGameObjectPool[spawn];

        if (obj->state == VACANT_SLOT)
            continue;

        // If the spawned object is outside of the unspawn window then it should be deactivated
        if (abs(obj->position.x-(camera.x+128)) > UNSPAWN_WINDOW_HORIZONTAL || abs(obj->position.y-(camera.y+128)) > UNSPAWN_WINDOW_VERTICAL)
        {
        	printf("UNSPAWN!\n");
			WaitKey();
            if (obj->state != GAMEOBJECT_STATE_INACTIVE)
                InactivateSpawnedGameObject(obj);

            obj->state = VACANT_SLOT;
            continue;
        }

        //Do AI. The AI routine takes care of setting object velocity as well.
        PerformAI(obj);

        //Set object displacement based on current velocity
		obj->displacement.x = obj->velocity.x;

		//This funky routine handles the jumping "physics"
		descending = obj->velocity.y < 0 ? true : false;
		if (abs(obj->velocity.y) >= (VERTICAL_VELOCITY_FACTOR*2)/3)
			obj->displacement.y = descending ? -2 : 2;
		else if (abs(obj->velocity.y) >= (VERTICAL_VELOCITY_FACTOR)/3)
			obj->displacement.y = obj->displacement.y = descending ? -1 : 1;
		else
			obj->displacement.y = 0;

		//Performa collision detection against world geometry based on current position and displacement.
        PerformStaticLevelCollision(obj);

		if (obj->displacement.x || obj->displacement.y)
		{
			//Move object sprite
			obj->position.x += obj->displacement.x;
			obj->position.y += obj->displacement.y;
			obj->updateRegisters = true;
		}

		if (obj->obj->type == GAMEOBJECT_TYPE_OURHERO)
			UpdateCamera(false);

		//Update Animation
		if (obj->obj->animations[obj->state]->numberOfFrames > 1 && ++obj->animationTimer % ANIMATION_UPDATE_INTERVAL == 0)
		{
			obj->currentFrame++;
			obj->updateRegisters = true;
		}

		if (obj->updateRegisters || camera.moved)
		{
			//This call takes care of updating the animation as well as positioning the sprite correctly according
			//to the current camera position.
			UpdateObjectAnimation(obj);
			obj->updateRegisters = false;
		}
		/*
		else if (camera.moved)
		{
			UpdateObjectPosition(obj);
		}
		*/
    }
	SetLayers(LAYER_SON_TON_GS0);
	wait_for_vdisp();
}

