#include "GameObject.h"
#include "animation.h"
#include "level.h"
#include "Camera.h"

void SpawnGameObject(short lx, short ly)
{
    short obj;
    SpawnedGameObject *gameObj;

    for (obj = 0; obj < MAX_SPAWNED_GAMEOBJECT_POOL_SIZE; obj++)
    {
        if (spawnedGameObjectPool[obj].state == VACANT_SLOT)
        {
            gameObj = &spawnedGameObjectPool[obj];

            gameObj->state = GAMEOBJECT_STATE_IDLE;
            gameObj->facing = GAMEOBJECT_FACING_RIGHT;
            gameObj->updateRegisters = false;
            gameObj->obj = &gameObjectPool[ObjectType(level.data[(ly << level.widthShiftFactor) + lx])];
            level.data[(ly << level.widthShiftFactor) + lx] = VACANT_SLOT;
            gameObj->origin.x = lx;
            gameObj->origin.y = ly;
            gameObj->position.x = lx << 4;
            gameObj->position.y = ly << 4;
            gameObj->displacement.x = 0;
            gameObj->displacement.y = 0;
            gameObj->velocity.x = 0;
            gameObj->velocity.y = 0;
            gameObj->currentFrame = 0;
            gameObj->animationTimer = 0;

			Wipe16(gameObj->spriteRegisters, -1, MAX_SPRITE_SEGMENTS);
            LoadSpriteIntoVRAMREGST(gameObj->obj->animations[GAMEOBJECT_STATE_IDLE]->frame[0], TranslateHorizontal(gameObj->position.x+gameObj->obj->horizontalSpriteOffset), TranslateVertical(gameObj->position.y+gameObj->obj->verticalSpriteOffset), false, gameObj->spriteRegisters);
            return;
        }
    }
}

//Spawns all the objects that have come within the spawn window around the visible screen area
//to avoid having to search the entire spawn window every frame we only attempt to spawn in the
//16x16 tiles that come into view whenever the camera moves 16 pixels in any direction and into another tile
void SpawnGameObjects(bool reset, short verticalChange, short horizontalChange)
{
	short x, y, v, h;
	VectorINT16 lvl;

	if (reset)
	{
		for (lvl.y = ((camera.y+128)>>4)-SPAWN_WINDOW_VERTICAL; lvl.y < SPAWN_WINDOW_VERTICAL<<1; lvl.y++)
		{
			//Skip areas that lie outside of the level boundary because there is nothing there to spawn.
			if (lvl.y < 0 || lvl.y >= level.height)
				continue;

			for (lvl.x = ((camera.x+128)>>4)-SPAWN_WINDOW_HORIZONTAL; lvl.x < SPAWN_WINDOW_HORIZONTAL<<1; lvl.x++)
			{
				//Skip areas that lie outside of the level boundary because there is nothing there to spawn.
				if (lvl.x < 0 || lvl.x >= level.width)
					continue;

				//if (level.data[]objects[lvl.y][lvl.x] != VACANT_SLOT)
				if (ObjectType(level.data[(lvl.y << level.widthShiftFactor) + lvl.x]) != VACANT_SLOT)
					SpawnGameObject(lvl.x, lvl.y);
			}
		}
		return;
	}

	//The camera has advanced vertically into a new 16x16 slot, so we need to spawn all objects
	//in the tiles that have moved into the spawn window
	if (verticalChange != 0)
	{
		if (verticalChange > 0)
			lvl.x = ((camera.x+128)>>4) + SPAWN_WINDOW_HORIZONTAL;
		else
			lvl.x = ((camera.x+128)>>4) - SPAWN_WINDOW_HORIZONTAL;

		if (lvl.x >= 0 && lvl.x < level.width)
		{
			for (lvl.y = ((camera.y+128)>>4) - SPAWN_WINDOW_VERTICAL; lvl.y < SPAWN_WINDOW_VERTICAL<<1; lvl.y++)
			{
				if (lvl.y < 0 || lvl.y >= level.height)
					continue;

				//if (level.objects[lvl.y][lvl.x] != VACANT_SLOT)
				if (ObjectType(level.data[(lvl.y << level.widthShiftFactor) + lvl.x]) != VACANT_SLOT)
					SpawnGameObject(lvl.x, lvl.y);
			}
		}
	}

	//The camera has advanced vertically into a new 16x16 slot, so we need to spawn all objects
	//in the tiles that have moved into the spawn window
	if (horizontalChange != 0)
	{
		if (horizontalChange > 0)
			lvl.y = ((camera.y+128)>>4) + SPAWN_WINDOW_VERTICAL;
		else
			lvl.y = ((camera.y+128)>>4) - SPAWN_WINDOW_VERTICAL;

		if (lvl.y >= 0 && lvl.y < level.height)
		{
			for (lvl.x = ((camera.x+128)>>4) - SPAWN_WINDOW_HORIZONTAL; lvl.x < SPAWN_WINDOW_HORIZONTAL<<1; lvl.x++)
			{
				if (lvl.x < 0 || lvl.x >= level.width)
					continue;

				//if (level.objects[lvl.y][lvl.x] != VACANT_SLOT)
				if (ObjectType(level.data[(lvl.y << level.widthShiftFactor) + lvl.x]) != VACANT_SLOT)
					SpawnGameObject(lvl.x, lvl.y);
			}
		}
	}
}

void InactivateSpawnedGameObject(SpawnedGameObject *gameObj)
{
    gameObj->state = GAMEOBJECT_STATE_INACTIVE;
    UnloadSpriteFromVRAMREGST(gameObj->spriteRegisters);
}

void UnspawnGameObject(SpawnedGameObject *gameObj)
{
    if (abs(TranslateHorizontal(gameObj->position.x)) > UNSPAWN_WINDOW_HORIZONTAL || abs(TranslateVertical(gameObj->position.y)) > UNSPAWN_WINDOW_VERTICAL)
        gameObj->state = VACANT_SLOT;
}

void inline UpdateObjectPosition(SpawnedGameObject *gameObj)
{
	int r, regst;

	for (r = 0; r < MAX_SPRITE_SEGMENTS; r++)
	{
		regst = gameObj->spriteRegisters[r];

		if (regst == -1)
			return;

		SREGST[regst].x += camera.dx;
		SREGST[regst].y += camera.dy;
	}
	/*
    short h, v, x, y;
    int regst, currentSegment = 0;
    GameSprite *gs = gameObj->obj->animations[gameObj->state]->frame[gameObj->currentFrame];

	x = TranslateHorizontal(gameObj->position.x);
	y = TranslateVertical(gameObj->position.y);

    for (v = 0; v < gs->verticalSegments; v++)
        for (h = 0; h < gs->horizontalSegments; h++, currentSegment++)
        {
            regst = spriteRegisters[currentSegment];
            SREGST[regst].x = x + (h << 4);
            SREGST[regst].y = y + (v << 4);
        }
	*/
}

