#ifndef __GAMEOBJECT_H__
#define __GAMEOBJECT_H__

#define MAX_OBJECT_ANIMATIONS               10
#define MAX_GAMEOBJECT_POOL_SIZE            100
#define MAX_SPAWNED_GAMEOBJECT_POOL_SIZE    30

#define GAMEOBJECT_STATE_IDLE               0
#define GAMEOBJECT_STATE_JUMPING           	1
#define GAMEOBJECT_STATE_RUNNING            2
#define GAMEOBJECT_STATE_ATTACKING          3
#define GAMEOBJECT_STATE_INACTIVE           254
#define GAMEOBJECT_FACING_RIGHT             0
#define GAMEOBJECT_FACING_LEFT              1

#define GAMEOBJECT_TYPE_OURHERO             0

#include "misc.h"
#include "Sprite.h"
//#include "animation.h"

typedef struct GameObject GameObject;
typedef struct Animation Animation;
typedef struct SpawnedGameObject SpawnedGameObject;

//typedef struct Animation;

//Game object is a master copy of any object that can be spawned in the game.
//there is only one copy of a GameObject but there can be many instances of
//the same GameObject via SpawnedGameObjects in the pool.
struct GameObject
{
    short type;
    short width;
    short height;
    short horizontalSpriteOffset;
    short verticalSpriteOffset;
    Animation *animations[MAX_OBJECT_ANIMATIONS];
} gameObjectPool[MAX_GAMEOBJECT_POOL_SIZE];

//This pool contain all the spawned game objects that currently exist
//basically this is a list of all currently active object in the game
struct SpawnedGameObject
{
    byte state;
    byte currentFrame;
    byte animationTimer;
    byte facing;
    bool updateRegisters;

    //The origin slot of object in "level.objects"
    //we need to remember this because we remove
    //the object from the level slot when we spawn
    //the object. If we want to reset the level we
    //need to put all spawned objects back into their
    //"level.objects" slots
    VectorINT16 origin;

    VectorINT16 position;
    VectorINT16 displacement;
    VectorINT16 velocity;

    GameObject *obj;
    short spriteRegisters[MAX_SPRITE_SEGMENTS];
} spawnedGameObjectPool[MAX_SPAWNED_GAMEOBJECT_POOL_SIZE];

void SpawnGameObject(short lx, short ly);
void SpawnGameObjects(bool reset, short verticalChange, short horizontalChange);
void InactivateSpawnedGameObject(SpawnedGameObject *gameObj);
void UnspawnGameObject(SpawnedGameObject *gameObj);
void inline UpdateObjectPosition(SpawnedGameObject *gameObj);

#endif
