#include "Level.h"
#include "Camera.h"
#include "Animation.h"
#include "GameObject.h"
#include "ResourceManager.h"
#include "Renderer.h"

int IsOnSolidSurface(SpawnedGameObject *obj)
{
	bool grounded = 0;
	short lx;
	short leftX = obj->position.x>>4;
	short rightX = (obj->position.x+obj->obj->width)>>4;
	short ly = (obj->position.y + obj->obj->height)>>4;

	if (ly < 0 || ly >= level.height)
		return;

	for (lx = leftX; lx <= rightX; lx++)
	{
		if (lx < 0 || lx >= level.width)
			continue;

		//if (level.geometry[ly][lx] != VACANT_SLOT)
		if (GeometryType(level.data[(ly << level.widthShiftFactor) + lx]) != VACANT_SLOT)
			grounded = 1;
	}
	return(grounded);
}

void PerformStaticLevelCollision(SpawnedGameObject *obj)
{
    short lx, ly;

    Box cBox;

	//This is the collision box of the object including its displacment
	//We use this to find a list of all tiles that the object might collide
	//with.
	cBox.x = obj->position.x;
	cBox.y = obj->position.y;
	cBox.width = obj->obj->width;
	cBox.height = obj->obj->height;

	if (obj->displacement.x > 0)
		cBox.width += obj->displacement.x;

	if (obj->displacement.x < 0)
	{
		cBox.x += obj->displacement.x;
		cBox.width += abs(obj->displacement.x);
	}

	if (obj->displacement.y > 0)
		cBox.height += obj->displacement.y;

	if (obj->displacement.y < 0)
	{
		cBox.y += obj->displacement.y;
		cBox.height += abs(obj->displacement.y);
	}

	for (ly = cBox.y >> 4; ly <= (cBox.y+cBox.height)>> 4; ly++)
		for (lx = cBox.x >> 4; lx <= (cBox.x+cBox.width)>> 4; lx++)
		{
			if (lx < 0 || ly < 0 || lx >= level.width || ly >= level.height)
				continue;

//			if (level.geometry[ly][lx] != VACANT_SLOT)
			if (GeometryType(level.data[(ly << level.widthShiftFactor) + lx]) != VACANT_SLOT)
			{
				CollideObjectLevelSolid(obj, lx<<4, ly<<4);
			}
		}
}

void CollideObjectLevelSolid(SpawnedGameObject *obj, short lx, short ly)
{
    short x, y;

    //If level block is contained inside the vertical range of obj then we should test
    //for horizontal collision
    y = obj->position.y + obj->displacement.y;
    if (!(ly+16 <= y || ly >= y + obj->obj->height))
    {
        //If object is moving to the right test for collision with the left side of 16x16 level block
        if (obj->displacement.x > 0)
        {
            //current momentum in x direction is causing a collision
            x = obj->position.x + obj->obj->width;
            if (x < lx && x + obj->displacement.x >= lx)
                obj->displacement.x = lx - x - 1;
        }

        //If object is moving to the left test for collision with the right side of 16x16 level block
        if (obj->displacement.x < 0)
        {
            //current momentum in x direction is causing a collision
            x = obj->position.x;
            if (x >= lx+16 && x + obj->displacement.x < lx+16)
				obj->displacement.x = (lx+16) - x;
        }
    }

    //If level block is contained inside the horizontal range of obj then we should test
    //for horizontal collision
    x = obj->position.x + obj->displacement.x;
    if (!(lx+16 <= x || lx >= x + obj->obj->width))
    {
        //If object is moving up test for collision with the bottom side of 16x16 level block
        if (obj->displacement.y > 0)
        {
            //current momentum in y direction is causing a landing collision
            y = obj->position.y + obj->obj->height;

            if (y < ly && y + obj->displacement.y >= ly)
				obj->displacement.y = ly - y;

			/*
            if (y <= ly && y + obj->displacement.y > ly)
            {
                obj->displacement.y = ly - y;
            }
            */
        }

        //If object is moving down test for collision with the upper side of 16x16 level block
        if (obj->displacement.y < 0)
        {
            //current momentum in y direction is causing a head bump collision
            y = obj->position.y;
            if (y >= ly+16 && y + obj->displacement.y < ly+16)
			{
				obj->displacement.y = (ly+16) - y;
				//Stricktly speaking we should only change object velocities in the AI routine
				//but for the sake of effeciency we take care it here.
				obj->velocity.x = 0;
				obj->velocity.y = 0;
			}
        }
    }
}

void LoadLevel(short levelNumber)
{
    char *levelName = "Level00.bin";
    char *bgName = "bg00.bin";
    short object, objects, x, y, ox, oy, type;
    short animation, frame;
    int fdLevel, fdBG;

    for (object = 0; object < MAX_SPAWNED_GAMEOBJECT_POOL_SIZE; object++)
        spawnedGameObjectPool[object].state = VACANT_SLOT;

    if (levelNumber%10 > 0)
	{
		levelName[6] = '0' + (levelNumber%10);
		bgName[3] = '0' + (levelNumber%10);
	}

    if (levelNumber/10 > 0)
	{
		levelName[5] = '0' + (levelNumber/10);
		bgName[2] = '0' + (levelNumber/10);
	}

    ResourceManager(RESOURCE_MANAGER_RESET);

    fdLevel = OpenFile(levelName, 0);

    level.width = FetchINT8(fdLevel);
    level.height = FetchINT8(fdLevel);

    if (level.width == 0)
		level.width = 256;

    if (level.height == 0)
		level.height = 256;

	level.widthShiftFactor = 1;

	while (1 << level.widthShiftFactor != level.width)
	{
		level.widthShiftFactor++;
	};

	level.data = AllocateMemory(level.width * level.height);

	//The two last bytes of level data pinpoint the begining point of the main character
	_dos_seek(fdLevel, (level.width * level.height) + 2, 0);
    ox = FetchINT8(fdLevel);
    oy = FetchINT8(fdLevel);

	//After fetching the coordinates of the main character we insert it into the level.
	level.data[(oy << level.widthShiftFactor) + ox] = GAMEOBJECT_TYPE_OURHERO + LEVEL_OBJECT_TYPE_OFFSET;

	//The offset of raw level data is 2 bytes into the level data since the forst two bytes are the
	//level dimensions
	_dos_seek(fdLevel,2,0);

	for (x = 0; x < level.width; x++)
		for (y = 0; y < level.height; y++)
		{
			type = FetchINT8(fdLevel);

			if (x != ox || y != oy)
				level.data[(y << level.widthShiftFactor) + x] = type;
			else
				type = level.data[(y << level.widthShiftFactor) + x];

			//Convert raw type to Object Type. If it is not an object then it is set equal to a vacant slot
			type = ObjectType(type);

			if (type != VACANT_SLOT)
			{
				//Load the animations for this game object into VRAM so it is ready to be pointed
				//to by sprite registers when object with said animations has been spawned.
				for (animation = 0; animation < MAX_OBJECT_ANIMATIONS; animation++)
					if (gameObjectPool[type].animations[animation] != null)
						for (frame = 0; frame < gameObjectPool[type].animations[animation]->numberOfFrames; frame++)
							LoadSpriteIntoVRAMPCG(gameObjectPool[type].animations[animation]->frame[frame]);


				//Because the camera initializes itself by focusing on the player character we need to spawn
				//the game character here.
				if (type == GAMEOBJECT_TYPE_OURHERO)
					SpawnGameObject(x, y);
			}
		}

    _dos_close(fdLevel);

    if (!CheckFile(bgName))
		backgroundSurface.enabled = false;
	else
	{
		backgroundSurface.enabled = true;

		fdBG = OpenFile(bgName, 0);

		background.width = FetchINT8(fdBG);
		background.height = FetchINT8(fdBG);

		if (background.width == 0)
			background.width = 256;

		if (background.height == 0)
			background.height = 256;

		//The Background scrolls at a pace of 1 to 2 compared to the foreground
		background.widthShiftFactor = 1;

		while (1 << background.widthShiftFactor != background.width)
		{
			background.widthShiftFactor++;
		};

		background.data = AllocateMemory(background.width * background.height);

		for (x = 0; x < background.width; x++)
			for (y = 0; y < background.height; y++)
			{
				type = FetchINT8(fdBG);
				background.data[(y << background.widthShiftFactor) + x] = type;
			}
	}

    _dos_close(fdBG);

	//Initial camera update and level render
	camera.type = CAMERA_TYPE_ALL_TRACE;
	UpdateCamera(true);

	if (levelSurface.enabled)
		RenderSurface(&levelSurface, true);

	if (backgroundSurface.enabled)
		RenderSurface(&backgroundSurface, true);
}
/*
void LoadLevel(short levelNumber)
{
    char *levelName = "Level00.dat";
    short object, objects, x, y, type;
    short animation, frame;
    int fdLevel;
    LevelFileData lfd;

    //Clear level
    Wipe32((void *)&level, 0xFFFFFFFF, sizeof(Level)>>2);
    for (object = 0; object < MAX_SPAWNED_GAMEOBJECT_POOL_SIZE; object++)
        spawnedGameObjectPool[object].state = VACANT_SLOT;

    if (levelNumber%10 > 0)
        levelName[6] = '0' + (levelNumber%10);
    if (levelNumber/10 > 0)
        levelName[5] = '0' + (levelNumber/10);

    ResourceManager(RESOURCE_MANAGER_RESET);

    fdLevel = OpenFile(levelName, 0);
    ReadFile(fdLevel, (void *)&lfd, sizeof(LevelFileData));
    _dos_seek(fdLevel,lfd.offsetLevelGameObjects, 0);

    //Fetch Level objects
    objects = FetchINT16(fdLevel);
    for (object = 0; object < objects; object++)
    {
        type = FetchINT16(fdLevel);
        x = FetchINT16(fdLevel)>>4;
        y = FetchINT16(fdLevel)>>4;
        level.objects[y][x] = type;

		//The player character will always be the first object to be spawned.
		if (type == GAMEOBJECT_TYPE_OURHERO)
			SpawnGameObject(x, y);

		//Load the animations for this game object into VRAM so it is ready to be pointed
		//to by sprite registers when object with said animations has been spawned.
        for (animation = 0; animation < MAX_OBJECT_ANIMATIONS; animation++)
            if (gameObjectPool[type].animations[animation] != null)
                for (frame = 0; frame < gameObjectPool[type].animations[animation]->numberOfFrames; frame++)
                    LoadSpriteIntoVRAMPCG(gameObjectPool[type].animations[animation]->frame[frame]);
    }

    //Fetch level geometry
    objects = FetchINT16(fdLevel);
    for (object = 0; object < objects; object++)
    {
        type = FetchINT16(fdLevel);
        x = FetchINT16(fdLevel);
        y = FetchINT16(fdLevel);
        level.geometry[y][x] = type;
    }

	//Initial camera update and level render
	camera.type = CAMERA_TYPE_ALL_TRACE;
	UpdateCamera(true);
	RenderLevel(true);

    _dos_close(fdLevel);
}
*/
/*
byte inline FetchMapData(TwoDimensionalMap *map, unsigned short x, unsigned short y)
{
	unsigned short offset = (y << map->widthShiftFactor) + x];
	byte data = map->date[offset];
	return(data);
}
*/
//To save space one byte of level data can contain either a level object or level geometry.
//All numbers from 0 up to just below LEVEL_OBJECT_TYPE_OFFSET represent level geometry, and
//all numbers equal to and above that represent level objects. 0xFF means the corresponding 16x16
//cell is empty
byte inline ObjectType(byte type)
{
	return((type >= LEVEL_OBJECT_TYPE_OFFSET && type != VACANT_SLOT) ? (type - LEVEL_OBJECT_TYPE_OFFSET) : VACANT_SLOT);
}

byte inline GeometryType(byte type)
{
	return(type < LEVEL_OBJECT_TYPE_OFFSET ? type : VACANT_SLOT);
}
