#include "Sprite.h"
#include "ResourceManager.h"
#include "GameObject.h"
//Allocates the necessary PCG's to contain whole of sprite and loads sprite into the PCG VRAM
void LoadSpriteIntoVRAMPCG(GameSprite *gs)
{
    byte pattern[128];
    int i, h, v, pcgIndex, currentSegment = 0;

    for (v = 0; v < gs->verticalSegments; v++)
        for (h = 0; h < gs->horizontalSegments; h++, currentSegment++)
        {
            pcgIndex = gs->segments[currentSegment]->pcgIndex;

            if (pcgIndex == VACANT_SLOT)
            {
                for (i = 0; i < 128; i++)
                    pattern[i] = gs->segments[currentSegment]->segment[i];

                pcgIndex = ResourceManager(RESOURCE_MANAGER_ALLOCATE_PCG);
                _iocs_sp_defcg(pcgIndex, 1, pattern);
            }
            gs->segments[currentSegment]->pcgIndex = pcgIndex;
        }
}

//If requested allocates necessary sprite registers and points them to sprite PCG's in order to make a sprite visible at a certain location.
void LoadSpriteIntoVRAMREGST(GameSprite *gs, short x, short y, bool flip, short *spriteRegisters)
{
	byte horizontalFlip = flip ? SP_HORIZONTAL_FLIP : 0;
    short h, v;
    int i, regst, currentSegment = 0;

    for (v = 0; v < gs->verticalSegments; v++)
        for (h = 0; h < gs->horizontalSegments; h++, currentSegment++)
        {
            if (spriteRegisters[currentSegment] == -1)
            {
                regst = ResourceManager(RESOURCE_MANAGER_ALLOCATE_REGST);
                spriteRegisters[currentSegment] = regst;
            }
            else regst = spriteRegisters[currentSegment];

            //_iocs_sp_regst(regst, 0, x + (h << 4), y + (v << 4), (gs->segments[currentSegment]->palletID<<8) | gs->segments[currentSegment]->pcgIndex, SP_PRIORITY_FRONT);
            if (flip)
				SREGST[regst].x = x + ((gs->horizontalSegments-1-h) << 4)+ (gs->horizontalSegments & 1 ? 0 : 16); //We add 16 pixels because sprites are offset by that much for some reason
			else
				SREGST[regst].x = x + (h << 4)+16;

            SREGST[regst].y = y + (v << 4)+16; //We add 16 pixels because sprites are offset by that much for some reason
            SREGST[regst].info = horizontalFlip | gs->segments[currentSegment]->palletID;
            SREGST[regst].id = gs->segments[currentSegment]->pcgIndex;
            SREGST[regst].priority = SP_PRIORITY_FRONT;
        }

        for (;currentSegment < MAX_SPRITE_SEGMENTS; currentSegment++)
			if (spriteRegisters[currentSegment] != -1)
			{
				regst = spriteRegisters[currentSegment];
				SREGST[regst].priority = SP_PRIORITY_INVISIBLE;
				resources.spriteRegister[regst] = RESOURCE_MANAGER_EMPTY_SLOT;
			}
}

void UnloadSpriteFromVRAMREGST(short *spriteRegisters)
{
	short index, regst;

	for (index = 0; index < MAX_SPRITE_SEGMENTS; index++)
		if (spriteRegisters[index] != -1)
		{
            regst = spriteRegisters[index];
			SREGST[regst].priority = SP_PRIORITY_INVISIBLE;
            resources.spriteRegister[regst] = RESOURCE_MANAGER_EMPTY_SLOT;
		}
}
