#include "misc.h"

void *AllocateMemory(int length)
{
	void *adr;

	if (heap == null)
	{
		heap =  _dos_malloc(HEAP_SIZE);
		#ifdef DEBUG
		if (heap >= (void *)0x81000000)
		{
			printf("Heap Allocation Error!\n");
			WaitKey();
		}
		#endif
	}

	//A negative allocation size means we get a temporary chunk of memory from the very end of the heap.
	//This should be used for things like loading the entirety of the sprite bitmap for processing.
	if (length < 0)
		adr = (void *)(heap + HEAP_SIZE + length - 1);
	else
	{
		adr = (void *)(heap + allocatedMemory);
		allocatedMemory += length;
	}

	#ifdef DEBUG
	if (allocatedMemory >= HEAP_SIZE)
	{
		printf("Heap too small Error!\n");
		WaitKey();
	}
	#endif
	/*
    void *adr = _dos_malloc(length);
    #ifdef DEBUG
    if (adr >= (void *)0x81000000)
    {
        printf("Allocation Error!\n");
        WaitKey();
    }
    #endif
    */
    return(adr);
}

int MakeFile(const char*file)
{
    int fd = _dos_create(file,0);
    #ifdef DEBUG
    if (fd < 0)
    {
        printf("File \"%s\" make Error!\n",file);
        WaitKey();
    }
    #endif

    return fd;
}

int CheckFile(const char*file)
{
    int fd = _dos_open(file,0);
    _dos_close(fd);

	if (fd < 0)
		return(0);
	else
		return(1);
}

int OpenFile(const char*file, int mode)
{
    int fd = _dos_open(file,mode);
    #ifdef DEBUG
    if (fd < 0)
    {
        printf("File \"%s\" open Error!\n", file);
        WaitKey();
    }
    //else printf("File \"%s\" opened!\n", file);
    #endif

    return fd;
}

void ReadFile(int fd, char *buffer, int length)
{
    int read = _dos_read(fd, buffer, length);
    #ifdef DEBUG
    if (read != length)
    {
        printf("File Read Error!\n");
        printf("Requested \"%i bytes\" but read \"%i bytes\"\n", length, read);
        WaitKey();
    }
    #endif
}

byte FetchINT8(int fd)
{
    byte data;
    ReadFile(fd, &data, 1);
    return(data);
}

short FetchINT16(int fd)
{
    short data;
    ReadFile(fd, (void *)&data, 2);
    return(data);
}

int FetchINT32(int fd)
{
    int data;
    ReadFile(fd, (void *) &data, 4);
    return(data);
}

void PushINT8(int fd, byte n)
{
    _dos_write(fd, (void *)&n, 1);
}

void PushINT16(int fd, short n)
{
    _dos_write(fd, (void *)&n, 2);
}

void PushINT32(int fd, int n)
{
    _dos_write(fd, (void *)&n, 4);
}

unsigned long EndianSwap32(unsigned long x)
{
    unsigned long y = 0;

    y |= (x&0x000000FF) << 24;
    y |= (x&0xFF000000) >> 24;
    y |= (x&0x0000FF00) << 8;
    y |= (x&0x00FF0000) >> 8;
    return y;
}

unsigned short EndianSwap16(unsigned short x)
{
    unsigned short y = 0;

    y |= (x&0x00FF) << 8;
    y |= (x&0xFF00) >> 8;
    return y;
}

void inline SetLayers(byte layers)
{
	*VIDEO_CONTROLLER_R2_ADR = layers;
	/*
    static short oldR2;
    short *R2 = VIDEO_CONTROLLER_R2_ADR;

    if (layers != LAYER_REVERT)
    {
        oldR2 = *R2;
        *R2 = layers;
    }
    else
    {
        *R2 = oldR2;
    }
    */
}

void Wipe16(short *ptr, short value, int length)
{
    int i;

    for (i = 0; i < length; i++)
        ptr[i] = value;
}

void Wipe32(int *ptr, int value, int length)
{
    int i;

    for (i = 0; i < length; i++)
        ptr[i] = value;
}

void CleanUp()
{
	//Remember to free heap allocation
	if (heap != null)
		_dos_mfree(heap);
	//Enable only the text layer
	SetLayers(LAYER_TON);
}

