#ifndef __MISC_H__
#define __MISC_H__

#ifdef DEBUG
#include <stdio.h>
#endif

#include <dos.h>
#include <iocs.h>

#define null                                ((void *)0x000000)
#define false                               0
#define true                                1
#define VACANT_SLOT                         0xFF

#define HEAP_SIZE							1048576

#define LAYER_NONE							0x00
#define LAYER_TON							0x20
#define LAYER_SON_GS0						0x41
#define LAYER_SON_TON_GS0					0x61
#define LAYER_SON_GS3_GS2_GS1_GS0           0x4F

//#define LAYER_REVERT                        0x80

#define SPRITE_REGISTER_ADR                 ((void *)0xEB0000)
#define VIDEO_CONTROLLER_R2_ADR             ((byte *)0xE82600)
#define SPRITE_PALLET_ADR                   ((unsigned short *)0xE82200)
#define GRAPHIC_PALLET_ADR                  ((unsigned short *)0xE82000)
#define MFP_BASE_ADR                        0xE88000
#define GVRAM0								((short *)0xC00000)
#define GVRAM1								((short *)0xC80000)
#define GVRAM0_POSITION						((void *)0xE80018) // GVRAM layer 0 position
#define GVRAM1_POSITION						((void *)0xE80020) // GVRAM layer 1 position

#define GPIP_ALARM							(1 << 0)
#define GPIP_EXPON							(1 << 1)
#define GPIP_POWSW							(1 << 2)
#define GPIP_OPMIRQ							(1 << 3)
#define GPIP_VDISP							(1 << 4)
#define GPIP_CRTC							(1 << 6)
#define GPIP_HSYNC							(1 << 7)

//#define GVRAMR12							((short *)0xE80018) // GVRAM layer 0 X position
//#define GVRAMR13							((short *)0xE8001A) // GVRAM layer 0 Y position
//#define GVRAMR14							((short *)0xE8001C) // GVRAM layer 1 X position
//#define GVRAMR15							((short *)0xE8001E) // GVRAM layer 1 Y position

#define mfp                                 ((*(volatile struct MFP *)MFP_BASE_ADR))

typedef unsigned char byte;
typedef unsigned short bool;
typedef struct VectorINT8 VectorINT8;
typedef struct VectorINT16 VectorINT16;

static void *heap = null;
static int allocatedMemory = 0;

struct VectorINT8
{
    char x;
    char y;
} __attribute__ ((packed));

struct VectorINT16
{
    short x;
    short y;
} __attribute__ ((packed));

/* MFP address */
struct MFP
 {
  unsigned short gpdr;
  unsigned short aer;
  unsigned short ddr;
  unsigned short iera;
  unsigned short ierb;
  unsigned short ipra;
  unsigned short iprb;
  unsigned short isra;
  unsigned short isrb;
  unsigned short imra;
  unsigned short imrb;
  unsigned short vr;
  unsigned short tacr;
  unsigned short tbcr;
  unsigned short tcdcr;
  unsigned short tadr;
  unsigned short tbdr;
  unsigned short tcdr;
  unsigned short tddr;
  unsigned short scr;
  unsigned short ucr;
  unsigned short rsr;
  unsigned short tsr;
  unsigned short udr;
} __attribute__ ((packed));

void *AllocateMemory(int length);
int MakeFile(const char*file);
int OpenFile(const char*file, int mode);
void ReadFile(int fd, char *buffer, int length);
byte FetchINT8(int fd);
short FetchINT16(int fd);
int FetchINT32(int fd);
void PushINT8(int fd, byte n);
void PushINT16(int fd, short n);
void PushINT32(int fd, int n);
unsigned long EndianSwap32(unsigned long x);
unsigned short EndianSwap16(unsigned short x);
void inline SetLayers(byte layers);
void Wipe16(short *ptr, short value, int length);
void Wipe32(int *ptr, int value, int length);

static void inline wait_for_vblank (void)
{
  /* CRTC under vdisp */
  while (mfp.gpdr & GPIP_VDISP);
}

static void inline wait_for_vdisp (void)
{
  /* CRTC under vblank */
  while (!(mfp.gpdr & GPIP_VDISP));
}
/*
static void inline wait_vblank(void)
{
  while (mfp.gpdr & 0xFF != 0x7A);
}
*/
static void inline WaitKey(void)
{
    _dos_inkey();
}

void CleanUp();

#endif

