#ifndef __SPRITE_H__
#define __SPRITE_H__

#define SP_PRIORITY_INVISIBLE               0x00
#define SP_PRIORITY_BACK                    0x01
#define SP_PRIORITY_MIDDLE                  0x02
#define SP_PRIORITY_FRONT                   0x03

#define SP_VERTICAL_FLIP                  	0x80
#define SP_HORIZONTAL_FLIP                  0x40

#define MAX_SPRITE_SEGMENTS                 20
#define MAX_SPRITE_POOL_SIZE                256
#define MAX_GRAPHIC_TILE_POOL_SIZE			50

#define SREGST								(((volatile struct SpriteData *)SPRITE_REGISTER_ADR))

#include "misc.h"

typedef struct SpriteData SpriteData;
typedef struct SpriteSegment SpriteSegment;
//typedef struct SpriteGraphic SpriteGraphic;
typedef struct GameSprite GameSprite;
typedef struct GraphicTile GraphicTile;

//This struct can be used to access the different members of the Sprite Data
//region of the X68000 directly when it points to the appropriate memory location.
struct SpriteData
{
    short x;
    short y;
    byte info;			//bit 0-4:High 4 bits of PALID, bit 6:HF, bit 7:VF
    byte id;
    short priority;	//0:Invisible, 1:Behind both BGs, 2:Behing bg0, 3:Front
} __attribute__ ((packed));

//This struct holds in system memory the graphical data for a 16x16 graphical segment
struct SpriteSegment
{
    byte palletID;
    byte pcgIndex;
    byte segment[128];
};

//This struct holds in system memory the graphical data for a 16x16 graphical tile
struct GraphicTile
{
    byte tile[256];
} graphicTilePool[MAX_GRAPHIC_TILE_POOL_SIZE];

//This struct holds information about an sprite image stored in main memory
struct GameSprite
{
    short graphicID;
    byte verticalSegments;
    byte horizontalSegments;
    //Holds information about segments that have been loaded into the VRAM from main memory
    SpriteSegment *segments[MAX_SPRITE_SEGMENTS];
} gameSpritePool[MAX_SPRITE_POOL_SIZE];

void LoadSpriteIntoVRAMPCG(GameSprite *gs);
void LoadSpriteIntoVRAMREGST(GameSprite *gs, short x, short y, bool flip, short *spriteRegisters);
//void inline ClearVRAMREGST(short regst);
void UnloadSpriteFromVRAMREGST(/*GameSprite *gs,*/ short *spriteRegisters);

#endif
